#
#   envelope.R
#
#   computes simulation envelopes 
#
#   $Revision: 1.45 $  $Date: 2008/05/11 20:21:08 $
#

envelope <- function(Y, fun=Kest, nsim=99, nrank=1, ..., 
                     simulate=NULL, verbose=TRUE, clipdata=TRUE, 
                     start=NULL, control=list(nrep=1e5, expand=1.5),
                     transform=NULL, global=FALSE, ginterval=NULL,
                     saveall=FALSE, nsim2=nsim, internal=NULL) {
  cl <- match.call()
  Yname <- deparse(substitute(Y))
  envir.user <- parent.frame()
  envir.here <- sys.frame(sys.nframe())
  # -------------------------------------------------------------------
  # Determine type of simulation
  #  simtype =
  #           "csr": uniform Poisson process
  #           "rmh": simulated realisation of fitted model 
  #          "expr": result of evaluating a user-supplied expression
  #          "list": user-supplied list of point patterns
  #
  # NB: **all** cases are handled in the code
  #     by setting up an expression `simexpr' that is repeatedly evaluated.
  #
  # Working variables
  #    'csr' = TRUE iff the model is (known to be) uniform Poisson
  #    model = NULL or a ppm object to be simulated
  #    envir = environment in which to evaluate the expression `simexpr'
  #
  if(is.null(simulate)) {
    # ...................................................
    # simulation type is determined by argument Y
    if(inherits(Y, "ppp")) {
      # Realisations of complete spatial randomness
      # will be generated by rpoispp
      simtype <- "csr"
      csr <- TRUE
      model <- NULL
      # Data pattern X is argument Y
      # Data pattern determines intensity of Poisson process
      X <- Y
      sY <- summary(Y, checkdup=FALSE)
      Yintens <- sY$intensity
      Ywin <- Y$window
      # expression that will be evaluated
      simexpr <- 
        if(!is.marked(Y)) {
          # unmarked point pattern
          expression(rpoispp(Yintens, win=Ywin))
        } else {
          # marked point pattern
          Ymarx <- marks(Y, dfok=FALSE)
          expression({A <- rpoispp(Yintens, win=Ywin);
                      A %mark% sample(Ymarx, A$n, replace=TRUE)})
        }
      envir    <- envir.here
    } else if(inherits(Y, "ppm")) {
      # Simulated realisations of the fitted model Y
      # will be generated using rmh
      simtype <- "rmh"
      csr <- FALSE
      model <- Y
      # Extract data pattern X from argument Y
      X <- data.ppm(model)
      # Set up parameters for rmh
      rmodel <- rmhmodel(model, verbose=FALSE)
      if(is.null(start))
        start <- list(n.start=X$n)
      rstart <- rmhstart(start)
      rcontr <- rmhcontrol(control)
      # pre-digest arguments
      rmhinfolist <- rmh(rmodel, rstart, rcontr, preponly=TRUE, verbose=FALSE)
      # expression that will be evaluated
      simexpr <- expression(rmhEngine(rmhinfolist, verbose=FALSE))
      envir <- envir.here
    } else 
    stop(paste(sQuote("Y"),
               "should be a point process model or a point pattern"))
  } else {
    # ...................................................
    # simulation is specified by argument `simulate'
    # which should be an expression, or a list of point patterns
    csr <- FALSE
    # override
    if(!is.null(internal$csr)) csr <- TRUE
    model <- NULL
    if(is.expression(simulate)) {
      # The user-supplied expression 'simulate' will be evaluated repeatedly
      simtype <- "expr"
      simexpr <- simulate
      envir <- envir.user
    } else if(is.list(simulate) &&
              all(unlist(lapply(simulate, inherits, what="ppp")))) {
      # The user-supplied list of point patterns will be used
      simtype <- "list"
      SimDataList <- simulate
      # expression that will be evaluated
      simexpr <- expression(SimDataList[[i]])
      envir <- envir.here
      # ensure that `i' is defined
      i <- 1 
    } else stop(paste(sQuote("simulate"),
                      "should be an expression, or a list of point patterns"))
    # extract data pattern X from argument Y
    if(inherits(Y, "ppp"))
      X <- Y
    else if(inherits(Y, "ppm"))
      X <- data.ppm(Y)
    else stop(paste(sQuote("Y"),
                    "should be a point process model or a point pattern"))
  }
  #--------------------------------------------------------------------
  # Determine number of simulations
  #
  #
  ## determine whether dual simulations are required
  ## (one set of simulations to calculate the theoretical mean,
  ##  another independent set of simulations to obtain the critical point.)
  dual <- (global && !csr)
  Nsim <- if(!dual) nsim else (nsim + nsim2)

  # if taking data from a list of point patterns,
  # check there are enough of them
  if(simtype == "list" && Nsim > length(SimDataList))
    stop(paste("Number of simulations",
               paren(if(!dual)
                     paste(nsim) else
                     paste(nsim, "+", nsim2, "=", Nsim)
                     ),
               "exceeds number of point pattern datasets supplied"))

  # Undocumented secret exit
  # ------------------------------------------------------------------
  if(!is.null(internal$patterns)) {
    # generate simulated realisations and return them 
    if(verbose) {
      action <- if(simtype == "list") "Extracting" else "Generating"
      descrip <- switch(simtype,
                        csr = "simulations of CSR",
                        rmh = "simulated realisations of fitted model",
                        expr = "simulations by evaluating expression",
                        list = "point patterns from list")
      cat(paste(action, Nsim, descrip, "...\n"))
    }
    XsimList <- list()
  # start simulation loop 
    for(i in 1:Nsim) {
      if(verbose) progressreport(i, Nsim)
      Xsim <- eval(simexpr, envir=envir)
      if(!inherits(Xsim, "ppp"))
        switch(simtype,
               csr=stop("Internal error: ppp object not generated"),
               rmh=stop("Internal error: rmh did not return a ppp object"),
               expr=stop(paste("Evaluating the expression", sQuote("simulate"),
                 "did not yield a point pattern")),
               list=stop("Internal error: list entry was not a ppp object"))
      XsimList[[i]] <- Xsim
    }
    if(verbose) cat(paste("Done.\n"))
    attr(XsimList, "internal") <- list(csr=csr)
    return(XsimList)
  }
  
  # ------------------------------------------------------------------
  # Summary function to be applied 
  
  # Name of function, for error messages
  fname <- if(is.name(substitute(fun))) deparse(substitute(fun))
           else if(is.character(fun)) fun else "fun"
  fname <- sQuote(fname)

  # R function to apply
  if(is.character(fun))
    fun <- get(fun)
  if(!is.function(fun)) 
    stop(paste("unrecognised format for function", fname))
  if(!any(c("r", "...") %in% names(formals(fun))))
    stop(paste(fname, "should have an argument",
               sQuote("r"), "or", sQuote("...")))

  # ---------------------------------------------------------------------
  # validate other arguments
  if((nrank %% 1) != 0)
    stop("nrank must be an integer")
  stopifnot(nrank > 0 && nrank < nsim/2)

  rgiven <- "r" %in% names(list(...))

  if(tran <- !is.null(transform)) {
    transform.funX <- eval(substitute(substitute(
                              eval(transform),
                              list("."=as.name("funX")))))
    transform.funXsim <- eval(substitute(substitute(
                              eval(transform),
                              list("."=as.name("funXsim")))))
  }
  if(!is.null(ginterval)) 
    stopifnot(is.numeric(ginterval) && length(ginterval) == 2)

  # rmh parameters are only relevant when simtype = "rmh"
  if(simtype != "rmh" && (!is.null(start) || !missing(control)))
      warning(paste("Poisson process simulation: arguments",
                    sQuote("start"),"and", sQuote("control"), "ignored"))

  # -------------------------------------------------------------------
  # Determine clipping window
  if(clipdata) {
    # Generate one realisation
    Xsim <- eval(simexpr, envir=envir)
    if(!inherits(Xsim, "ppp"))
      switch(simtype,
             csr=stop("Internal error: ppp object not generated"),
             rmh=stop("Internal error: rmh did not return a ppp object"),
             expr=stop(paste("Evaluating the expression", sQuote("simulate"),
                   "did not yield a point pattern")),
             list=stop("Internal error: list entry was not a ppp object"))
    # Extract window
    clipwin <- Xsim$window
    if(!is.subset.owin(clipwin, X$window))
      warning("Window containing simulated patterns is not a subset of data window")
  }
  
  # ---------------------------------------------------------------------
  # Evaluate function for data pattern X 
  if(!clipdata)
    funX <- fun(X, ...)
  else
    funX <- fun(X[clipwin], ...)

  if(!inherits(funX, "fv"))
    stop(paste("The function", fname,
               "must return an object of class", sQuote("fv")))

  argname <- attr(funX, "argu")
  valname <- attr(funX, "valu")

  if(tran) {
    # extract only the recommended value
    if(csr) 
      funX <- funX[, c(argname, valname, "theo")]
    else
      funX <- funX[, c(argname, valname)]
    # apply the transformation to it
    funX <- do.call("eval.fv", list(transform.funX))
  }

  rvals <- funX[[argname]]
  fX    <- funX[[valname]]

  # default domain over which to maximise
  alim <- attr(funX, "alim")
  if(global && is.null(ginterval))
    ginterval <- alim

  
  ######### simulate #######################
  if(verbose) {
    action <- if(simtype == "list") "Extracting" else "Generating"
    descrip <- switch(simtype,
                      csr = "simulations of CSR",
                      rmh = "simulated realisations of fitted model",
                      expr = "simulations by evaluating expression",
                      list = "point patterns from list")
    cat(paste(action, Nsim, descrip, "...\n"))
  }
  nrvals <- length(rvals)
  simvals <- matrix(, nrow=nrvals, ncol=Nsim)

  # start simulation loop 
  for(i in 1:Nsim) {
    Xsim <- eval(simexpr, envir=envir)
    if(!inherits(Xsim, "ppp"))
      switch(simtype,
             csr=stop("Internal error: ppp object not generated"),
             rmh=stop("Internal error: rmh did not return a ppp object"),
             expr=stop(paste("Evaluating the expression", sQuote("simulate"),
                   "did not yield a point pattern")),
             list=stop("Internal error: list entry was not a ppp object"))
    
    # apply function
    if(rgiven) 
      funXsim <- fun(Xsim, ...)
    else
      funXsim <- fun(Xsim, r=rvals, ...)
    # sanity checks
    if(!inherits(funXsim, "fv"))
      stop(paste("When applied to a simulated pattern, the function",
                 fname, "did not return an object of class",
                 sQuote("fv")))
    argname.sim <- attr(funXsim, "argu")
    valname.sim <- attr(funXsim, "valu")
    if(argname.sim != argname)
      stop(paste("The objects returned by", fname,
                 "when applied to a simulated pattern",
                 "and to the data pattern",
                 "are incompatible. They have different argument names",
                 sQuote(argname.sim), "and", sQuote(argname), 
                 "respectively"))
    if(valname.sim != valname)
      stop(paste("When", fname, "is applied to a simulated pattern",
                 "it provides an estimate named", sQuote(valname.sim), 
                 "whereas the estimate for the data pattern is named",
                 sQuote(valname),
                 ". Try using the argument", sQuote("correction"),
                 "to make them compatible"))

    if(tran) {
      # extract only the recommended value
      if(csr) 
        funXsim <- funXsim[, c(argname, valname, "theo")]
      else
        funXsim <- funXsim[, c(argname, valname)]
      # apply the transformation to it
      funXsim <- do.call("eval.fv", list(transform.funXsim))
    }
      
    # extract the values for simulation i
    simvals.i <- funXsim[[valname]]
    if(length(simvals.i) != nrvals)
      stop("Vectors of function values have incompatible lengths")
      
    simvals[ , i] <- funXsim[[valname]]
    if(verbose)
      progressreport(i, Nsim)
  }
  ##  end simulation loop
  
  if(verbose)
    cat("\nDone.\n")

  # compute envelopes
  orderstat <- function(x, n) { sort(x)[n] }
  if(!global) {
    # POINTWISE ENVELOPES
    lo <- apply(simvals, 1, orderstat, n=nrank)
    hi <- apply(simvals, 1, orderstat, n=nsim-nrank+1)
    lo.name <- paste("lower pointwise envelope of simulations")
    hi.name <- paste("upper pointwise envelope of simulations")

    if(csr) {
      results <- data.frame(r=rvals,
                            obs=fX,
                            theo=funX[["theo"]],
                            lo=lo,
                            hi=hi)
    } else {
      m <- apply(simvals, 1, mean, na.rm=TRUE)
      results <- data.frame(r=rvals,
                            obs=fX,
                            mmean=m,
                            lo=lo,
                            hi=hi)
    }
  } else {
    # SIMULTANEOUS ENVELOPES
    domain <- (rvals >= ginterval[1]) & (rvals <= ginterval[2])
    funX <- funX[domain, ]
    simvals <- simvals[domain, ]
    if(csr)
      theory <- funX[["theo"]]
    else {
      # use the first 'nsim' results to estimate the mean under H0
      theory <- m <- apply(simvals[, 1:nsim], 1, mean, na.rm=TRUE)
      # then discard them and use the remaining 'nsim2' simulations
      # to construct the envelopes.
      simvals <- simvals[, nsim + 1:nsim2]
    }
    # compute max absolute deviations
    deviations <- sweep(simvals, 1, theory)
    suprema <- apply(abs(deviations), 2, max, na.rm=TRUE)
    # ranked deviations
    dmax <- orderstat(suprema, n=nsim-nrank+1)
    # simultaneous bands
    lo <- theory - dmax
    hi <- theory + dmax
    lo.name <- "lower critical boundary"
    hi.name <- "upper critical boundary"

    if(csr)
      results <- data.frame(r=rvals[domain],
                            obs=fX[domain],
                            theo=theory,
                            lo=lo,
                            hi=hi)
    else
      results <- data.frame(r=rvals[domain],
                            obs=fX[domain],
                            mmean=m,
                            lo=lo,
                            hi=hi)
  }
  
  result <- fv(results,
               argu="r",
               ylab=attr(funX, "ylab"),
               valu="obs",
               fmla= deparse(. ~ r),
               alim=attr(funX, "alim"),
               labl=c("r", "obs(r)", if(csr) "theo(r)" else "mean(r)",
                 "lo(r)", "hi(r)"),
               desc=c("distance argument r",
                 "function value for data pattern",
                 if(csr) "theoretical value for CSR"
                 else "mean of simulations",
                 lo.name, hi.name))
  attr(result, "dotnames") <- c("obs",
                                if(csr) "theo" else "mmean",
                                "hi", "lo")
  unitname(result) <- unitname(funX)

  class(result) <- c("envelope", class(result))
  attr(result, "einfo") <- list(call=cl,
                                Yname=Yname,
                                csr=csr,
                                simtype=simtype,
                                nrank=nrank,
                                nsim=nsim,
                                global=global,
                                dual=dual,
                                nsim2=nsim2)
  if(saveall) {
    alldata <- cbind(rvals, simvals)
    simnames <- paste("sim", 1:nsim, sep="")
    colnames(alldata) <- c("r", simnames)
    alldata <- as.data.frame(alldata)
    savedata <- fv(alldata,
                   argu="r",
                   ylab=attr(funX, "ylab"),
                   valu="sim1",
                   fmla= deparse(. ~ r),
                   alim=attr(funX, "alim"),
                   labl=names(alldata),
                   desc=c("distance argument r",
                     paste("Simulation ", 1:nsim, sep="")))
    attr(savedata, "dotnames") <- simnames
    attr(result, "savedata") <- savedata
  }
  return(result)
}


print.envelope <- function(x, ...) {
  e <- attr(x, "einfo")
  g <- e$global
  csr <- e$csr
  simtype <- e$simtype
  nr <- e$nrank
  nsim <- e$nsim
  fname <- deparse(attr(x, "ylab"))
  type <- if(g) "Simultaneous" else "Pointwise"
  cat(paste(type, "critical envelopes for", fname, "\n"))
  descrip <-
    if(!is.null(simtype)) {
      switch(simtype,
             csr="simulations of CSR",
             rmh="simulations of fitted model",
             expr="evaluations of user-supplied expression",
             list="point pattern datasets in user-supplied list")
    } else if(csr) "CSR" else "model"
  #  
  cat(paste("Obtained from", nsim,
            "simulations of", descrip, "\n"))
  if(!is.null(e$dual) && e$dual) 
    cat(paste("Theoretical (i.e. null) mean value of", fname,
              "estimated from a separate set of",
              e$nsim2, "simulations\n"))
  if(!is.null(attr(x, "savedata"))) 
    cat("(All simulated values are stored)\n")
  alpha <- if(g) { nr/(nsim+1) } else { 2 * nr/(nsim+1) }
  cat(paste("Significance level of",
            if(!g) "pointwise",
            "Monte Carlo test:",
            paste(if(g) nr else 2 * nr,
                  "/", nsim+1, sep=""),
            "=", alpha, "\n"))
  cat(paste("Data:", e$Yname, "\n"))
  NextMethod("print")
}
                  
summary.envelope <- function(object, ...) {
  e <- attr(object, "einfo")
  g <- e$global
  nr <- e$nrank
  nsim <- e$nsim
  csr <- e$csr
  simtype <- e$simtype
  fname <- deparse(attr(object, "ylab"))
  type <- if(g) "Simultaneous" else "Pointwise"
  cat(paste(type, "critical envelopes for", fname, "\n"))
  descrip <-
    if(!is.null(simtype)) {
      switch(simtype,
             csr="simulations of CSR",
             rmh="simulations of fitted model",
             expr="evaluations of user-supplied expression",
             list="point pattern datasets in user-supplied list")
    } else if(csr) "CSR" else "model"
  #  
  cat(paste("Obtained from", nsim,
            "simulations of", descrip, "\n"))
  if(!is.null(attr(object, "savedata")))
    cat(paste("(All", nsim, "simulated values",
              "are stored in attr(,", dQuote("savedata"), ") )\n"))
  ordinal <- function(k) {
    last <- k %% 10
    ending <- if(last == 1) "st" else
              if(last == 2) "nd" else
              if(last== 3) "rd" else "th"
    return(paste(k, ending, sep=""))
  }
  lo.ord <- if(nr == 1) "minimum" else paste(ordinal(nr), "smallest")
  hi.ord <- if(nr == 1) "maximum" else paste(ordinal(nsim-nr+1), "largest")
  if(g) 
    cat(paste("Envelopes computed as",
              if(csr) "theoretical curve" else "mean of simulations",
              "plus/minus", hi.ord,
              "simulated value of maximum absolute deviation\n"))
  else {
    cat(paste("Upper envelope: pointwise", hi.ord,"of simulated curves\n"))
    cat(paste("Lower envelope: pointwise", lo.ord,"of simulated curves\n"))
  }
  alpha <- if(g) { nr/(nsim+1) } else { 2 * nr/(nsim+1) }
  cat(paste("Significance level of Monte Carlo test:",
            paste(if(g) nr else 2 * nr,
                  "/", nsim+1, sep=""),
            "=", alpha, "\n"))
  cat(paste("Data:", e$Yname, "\n"))
  return(invisible(NULL))
}
  
  
  
  
